@@ -8,9 +8,13 @@ Gtk.init(null);
 let win = new Gtk.Window();
 
 let view = new WebKit.WebView();
-view.load_uri("http://www.google.com/");
+view.load_uri('http://www.google.com/');
 win.add(view);
 
+win.connect('destroy', () => {
+    Gtk.main_quit();
+});
+
 win.set_size_request(640, 480);
 win.show_all();
 
