@@ -23,12 +23,17 @@
 
 #include <config.h>
 
-#include <string.h>
-#include <girepository.h>
+#include <string.h>  // for strcmp
 
-#include "arg.h"
-#include "foreign.h"
-#include "cjs/jsapi-wrapper.h"
+#include <girepository.h>
+#include <glib.h>
+
+#include <js/TypeDecls.h>
+
+#include "gi/arg.h"
+#include "gi/foreign.h"
+#include "cjs/context-private.h"
+#include "cjs/jsapi-util.h"
 
 static struct {
     char *gi_namespace;
@@ -41,9 +46,7 @@ static struct {
 
 static GHashTable* foreign_structs_table = NULL;
 
-static GHashTable*
-get_foreign_structs(void)
-{
+[[nodiscard]] static GHashTable* get_foreign_structs() {
     // FIXME: look into hasing on GITypeInfo instead.
     if (!foreign_structs_table) {
         foreign_structs_table = g_hash_table_new_full(g_str_hash, g_str_equal,
@@ -54,15 +57,11 @@ get_foreign_structs(void)
     return foreign_structs_table;
 }
 
-static bool
-gjs_foreign_load_foreign_module(JSContext *context,
-                                const gchar *gi_namespace)
-{
+[[nodiscard]] static bool gjs_foreign_load_foreign_module(
+    JSContext* context, const char* gi_namespace) {
     int i;
 
     for (i = 0; foreign_modules[i].gi_namespace; ++i) {
-        int code;
-        GError *error = NULL;
         char *script;
 
         if (strcmp(gi_namespace, foreign_modules[i].gi_namespace) != 0)
@@ -74,12 +73,11 @@ gjs_foreign_load_foreign_module(JSContext *context,
         // FIXME: Find a way to check if a module is imported
         //        and only execute this statement if isn't
         script = g_strdup_printf("imports.%s;", gi_namespace);
-        if (!gjs_context_eval((GjsContext*) JS_GetContextPrivate(context), script, strlen(script),
-                              "<internal>", &code,
-                              &error)) {
-            g_printerr("ERROR: %s\n", error->message);
+        JS::RootedValue retval(context);
+        GjsContextPrivate* gjs = GjsContextPrivate::from_cx(context);
+        if (!gjs->eval_with_scope(nullptr, script, -1, "<internal>", &retval)) {
+            g_critical("ERROR importing foreign module %s\n", gi_namespace);
             g_free(script);
-            g_error_free(error);
             return false;
         }
         g_free(script);
@@ -90,26 +88,20 @@ gjs_foreign_load_foreign_module(JSContext *context,
     return false;
 }
 
-bool
-gjs_struct_foreign_register(const char *gi_namespace,
-                            const char *type_name,
-                            GjsForeignInfo *info)
-{
+void gjs_struct_foreign_register(const char* gi_namespace,
+                                 const char* type_name, GjsForeignInfo* info) {
     char *canonical_name;
 
-    g_return_val_if_fail(info != NULL, false);
-    g_return_val_if_fail(info->to_func != NULL, false);
-    g_return_val_if_fail(info->from_func != NULL, false);
+    g_return_if_fail(info);
+    g_return_if_fail(info->to_func);
+    g_return_if_fail(info->from_func);
 
     canonical_name = g_strdup_printf("%s.%s", gi_namespace, type_name);
     g_hash_table_insert(get_foreign_structs(), canonical_name, info);
-    return true;
 }
 
-static GjsForeignInfo *
-gjs_struct_foreign_lookup(JSContext  *context,
-                          GIBaseInfo *interface_info)
-{
+[[nodiscard]] static GjsForeignInfo* gjs_struct_foreign_lookup(
+    JSContext* context, GIBaseInfo* interface_info) {
     GjsForeignInfo *retval = NULL;
     GHashTable *hash_table;
     char *key;
