@@ -770,14 +770,21 @@ bool ObjectBase::id_is_never_lazy(jsid name, const GjsAtoms& atoms) {
 }
 
 bool ObjectPrototype::resolve_impl(JSContext* context, JS::HandleObject obj,
-                                   JS::HandleId id, const char* name,
-                                   bool* resolved) {
+                                   JS::HandleId id, bool* resolved) {
     if (m_unresolvable_cache.has(id)) {
         *resolved = false;
         return true;
     }
 
-    if (!uncached_resolve(context, obj, id, name, resolved))
+    JS::UniqueChars prop_name;
+    if (!gjs_get_string_id(context, id, &prop_name))
+        return false;
+    if (!prop_name) {
+        *resolved = false;
+        return true;  // not resolved, but no error
+    }
+
+    if (!uncached_resolve(context, obj, id, prop_name.get(), resolved))
         return false;
 
     if (!*resolved && !m_unresolvable_cache.putNew(id)) {
@@ -1256,13 +1263,16 @@ static void wrapped_gobj_toggle_notify(void*, GObject* gobj,
          */
         if (is_main_thread) {
             if (G_UNLIKELY (toggle_up_queued || toggle_down_queued)) {
-                g_error("toggling down object %s that's already queued to toggle %s\n",
-                        G_OBJECT_TYPE_NAME(gobj),
+                g_critical("toggling down object %s (%p) that's already queued to toggle %s\n",
+                        G_OBJECT_TYPE_NAME(gobj), gobj,
                         toggle_up_queued && toggle_down_queued? "up and down" :
                         toggle_up_queued? "up" : "down");
+
+                toggle_queue.enqueue(gobj, ToggleQueue::DOWN, toggle_handler);
+            } else {
+                ObjectInstance::for_gobject(gobj)->toggle_down();
             }
 
-            ObjectInstance::for_gobject(gobj)->toggle_down();
         } else {
             toggle_queue.enqueue(gobj, ToggleQueue::DOWN, toggle_handler);
         }
