@@ -1,14 +1,29 @@
-#include "cjs/jsapi-util.h"
+#include <config.h>
+
+#include <glib.h>
+
+#include <js/Class.h>
+#include <js/GCAPI.h>  // for JS_GC, JS_SetGCCallback, JSGCStatus
+#include <js/RootingAPI.h>
+#include <js/TypeDecls.h>
+#include <js/Value.h>
+#include <jsapi.h>  // for JS_GetPrivate, JS_NewObject, JS_Set...
+
 #include "cjs/jsapi-util-root.h"
-#include "gjs-test-utils.h"
+#include "test/gjs-test-utils.h"
+
+// COMPAT: https://gitlab.gnome.org/GNOME/glib/-/merge_requests/1553
+#ifdef __clang_analyzer__
+void g_assertion_message(const char*, const char*, int, const char*,
+                         const char*) __attribute__((analyzer_noreturn));
+#endif
 
 static GMutex gc_lock;
 static GCond gc_finished;
 static volatile int gc_counter;
 
 #define PARENT(fx) ((GjsUnitTestFixture *)fx)
-typedef struct _GjsRootingFixture GjsRootingFixture;
-struct _GjsRootingFixture {
+struct GjsRootingFixture {
     GjsUnitTestFixture parent;
 
     bool finalized;
@@ -17,25 +32,20 @@ struct _GjsRootingFixture {
     GjsMaybeOwned<JSObject *> *obj;  /* only used in callback test cases */
 };
 
-static void
-test_obj_finalize(JSFreeOp *fop,
-                  JSObject *obj)
-{
+static void test_obj_finalize(JSFreeOp*, JSObject* obj) {
     bool *finalized_p = static_cast<bool *>(JS_GetPrivate(obj));
     g_assert_false(*finalized_p);
     *finalized_p = true;
 }
 
 static const JSClassOps test_obj_class_ops = {
-    NULL,  /* addProperty */
-    NULL,  /* deleteProperty */
-    NULL,  /* getProperty */
-    NULL,  /* setProperty */
-    NULL,  /* enumerate */
-    NULL,  /* resolve */
-    nullptr,  /* mayResolve */
-    test_obj_finalize
-};
+    nullptr,  // addProperty
+    nullptr,  // deleteProperty
+    nullptr,  // enumerate
+    nullptr,  // newEnumerate
+    nullptr,  // resolve
+    nullptr,  // mayResolve
+    test_obj_finalize};
 
 static JSClass test_obj_class = {
     "TestObj",
@@ -51,11 +61,7 @@ test_obj_new(GjsRootingFixture *fx)
     return retval;
 }
 
-static void
-on_gc(JSContext *cx,
-      JSGCStatus status,
-      void      *data)
-{
+static void on_gc(JSContext*, JSGCStatus status, JS::GCReason, void*) {
     if (status != JSGC_END)
         return;
 
@@ -94,30 +100,24 @@ wait_for_gc(GjsRootingFixture *fx)
     g_mutex_unlock(&gc_lock);
 }
 
-static void
-test_maybe_owned_rooted_flag_set_when_rooted(GjsRootingFixture *fx,
-                                             gconstpointer      unused)
-{
+static void test_maybe_owned_rooted_flag_set_when_rooted(GjsRootingFixture* fx,
+                                                         const void*) {
     auto obj = new GjsMaybeOwned<JS::Value>();
     obj->root(PARENT(fx)->cx, JS::TrueValue());
     g_assert_true(obj->rooted());
     delete obj;
 }
 
-static void
-test_maybe_owned_rooted_flag_not_set_when_not_rooted(GjsRootingFixture *fx,
-                                                     gconstpointer      unused)
-{
+static void test_maybe_owned_rooted_flag_not_set_when_not_rooted(
+    GjsRootingFixture*, const void*) {
     auto obj = new GjsMaybeOwned<JS::Value>();
     *obj = JS::TrueValue();
     g_assert_false(obj->rooted());
     delete obj;
 }
 
-static void
-test_maybe_owned_rooted_keeps_alive_across_gc(GjsRootingFixture *fx,
-                                              gconstpointer      unused)
-{
+static void test_maybe_owned_rooted_keeps_alive_across_gc(GjsRootingFixture* fx,
+                                                          const void*) {
     auto obj = new GjsMaybeOwned<JSObject *>();
     obj->root(PARENT(fx)->cx, test_obj_new(fx));
 
@@ -129,10 +129,8 @@ test_maybe_owned_rooted_keeps_alive_across_gc(GjsRootingFixture *fx,
     g_assert_true(fx->finalized);
 }
 
-static void
-test_maybe_owned_rooted_is_collected_after_reset(GjsRootingFixture *fx,
-                                                 gconstpointer      unused)
-{
+static void test_maybe_owned_rooted_is_collected_after_reset(
+    GjsRootingFixture* fx, const void*) {
     auto obj = new GjsMaybeOwned<JSObject *>();
     obj->root(PARENT(fx)->cx, test_obj_new(fx));
     obj->reset();
@@ -142,22 +140,28 @@ test_maybe_owned_rooted_is_collected_after_reset(GjsRootingFixture *fx,
     delete obj;
 }
 
-static void
-test_maybe_owned_weak_pointer_is_collected_by_gc(GjsRootingFixture *fx,
-                                                 gconstpointer      unused)
-{
+static void update_weak_pointer(JSContext*, JS::Compartment*, void* data) {
+    auto* obj = static_cast<GjsMaybeOwned<JSObject*>*>(data);
+    if (*obj)
+        obj->update_after_gc();
+}
+
+static void test_maybe_owned_weak_pointer_is_collected_by_gc(
+    GjsRootingFixture* fx, const void*) {
     auto obj = new GjsMaybeOwned<JSObject *>();
     *obj = test_obj_new(fx);
 
+    JS_AddWeakPointerCompartmentCallback(PARENT(fx)->cx, &update_weak_pointer,
+                                         obj);
     wait_for_gc(fx);
     g_assert_true(fx->finalized);
+    JS_RemoveWeakPointerCompartmentCallback(PARENT(fx)->cx,
+                                            &update_weak_pointer);
     delete obj;
 }
 
-static void
-test_maybe_owned_heap_rooted_keeps_alive_across_gc(GjsRootingFixture *fx,
-                                                   gconstpointer      unused)
-{
+static void test_maybe_owned_heap_rooted_keeps_alive_across_gc(
+    GjsRootingFixture* fx, const void*) {
     auto obj = new GjsMaybeOwned<JSObject *>();
     obj->root(PARENT(fx)->cx, test_obj_new(fx));
 
@@ -169,10 +173,8 @@ test_maybe_owned_heap_rooted_keeps_alive_across_gc(GjsRootingFixture *fx,
     g_assert_true(fx->finalized);
 }
 
-static void
-test_maybe_owned_switching_mode_keeps_same_value(GjsRootingFixture *fx,
-                                                 gconstpointer      unused)
-{
+static void test_maybe_owned_switching_mode_keeps_same_value(
+    GjsRootingFixture* fx, const void*) {
     JSObject *test_obj = test_obj_new(fx);
     auto obj = new GjsMaybeOwned<JSObject *>();
 
@@ -183,17 +185,15 @@ test_maybe_owned_switching_mode_keeps_same_value(GjsRootingFixture *fx,
     g_assert_true(obj->rooted());
     g_assert_true(*obj == test_obj);
 
-    obj->switch_to_unrooted();
+    obj->switch_to_unrooted(PARENT(fx)->cx);
     g_assert_false(obj->rooted());
     g_assert_true(*obj == test_obj);
 
     delete obj;
 }
 
-static void
-test_maybe_owned_switch_to_rooted_prevents_collection(GjsRootingFixture *fx,
-                                                      gconstpointer      unused)
-{
+static void test_maybe_owned_switch_to_rooted_prevents_collection(
+    GjsRootingFixture* fx, const void*) {
     auto obj = new GjsMaybeOwned<JSObject *>();
     *obj = test_obj_new(fx);
 
@@ -204,24 +204,23 @@ test_maybe_owned_switch_to_rooted_prevents_collection(GjsRootingFixture *fx,
     delete obj;
 }
 
-static void
-test_maybe_owned_switch_to_unrooted_allows_collection(GjsRootingFixture *fx,
-                                                      gconstpointer      unused)
-{
+static void test_maybe_owned_switch_to_unrooted_allows_collection(
+    GjsRootingFixture* fx, const void*) {
     auto obj = new GjsMaybeOwned<JSObject *>();
     obj->root(PARENT(fx)->cx, test_obj_new(fx));
 
-    obj->switch_to_unrooted();
+    obj->switch_to_unrooted(PARENT(fx)->cx);
+    JS_AddWeakPointerCompartmentCallback(PARENT(fx)->cx, &update_weak_pointer,
+                                         obj);
     wait_for_gc(fx);
     g_assert_true(fx->finalized);
+    JS_RemoveWeakPointerCompartmentCallback(PARENT(fx)->cx,
+                                            &update_weak_pointer);
 
     delete obj;
 }
 
-static void
-context_destroyed(JS::HandleObject obj,
-                  void            *data)
-{
+static void context_destroyed(JS::HandleObject, void* data) {
     auto fx = static_cast<GjsRootingFixture *>(data);
     g_assert_false(fx->notify_called);
     g_assert_false(fx->finalized);
@@ -229,10 +228,8 @@ context_destroyed(JS::HandleObject obj,
     fx->obj->reset();
 }
 
-static void
-test_maybe_owned_notify_callback_called_on_context_destroy(GjsRootingFixture *fx,
-                                                           gconstpointer      unused)
-{
+static void test_maybe_owned_notify_callback_called_on_context_destroy(
+    GjsRootingFixture* fx, const void*) {
     fx->obj = new GjsMaybeOwned<JSObject *>();
     fx->obj->root(PARENT(fx)->cx, test_obj_new(fx), context_destroyed, fx);
 
@@ -241,10 +238,8 @@ test_maybe_owned_notify_callback_called_on_context_destroy(GjsRootingFixture *fx
     delete fx->obj;
 }
 
-static void
-test_maybe_owned_object_destroyed_after_notify(GjsRootingFixture *fx,
-                                               gconstpointer      unused)
-{
+static void test_maybe_owned_object_destroyed_after_notify(
+    GjsRootingFixture* fx, const void*) {
     fx->obj = new GjsMaybeOwned<JSObject *>();
     fx->obj->root(PARENT(fx)->cx, test_obj_new(fx), context_destroyed, fx);
 
@@ -256,8 +251,9 @@ test_maybe_owned_object_destroyed_after_notify(GjsRootingFixture *fx,
 void
 gjs_test_add_tests_for_rooting(void)
 {
-#define ADD_ROOTING_TEST(path, f) \
-    g_test_add("/rooting/" path, GjsRootingFixture, NULL, setup, f,  teardown);
+#define ADD_ROOTING_TEST(path, f)                                      \
+    g_test_add("/rooting/" path, GjsRootingFixture, nullptr, setup, f, \
+               teardown);
 
     ADD_ROOTING_TEST("maybe-owned/rooted-flag-set-when-rooted",
                      test_maybe_owned_rooted_flag_set_when_rooted);
